// -*- C++ -*-
/*!
 * @file  RobotGPSDataManager.h
 * @brief gps data management
 * @date  $Date$
 *
 * @author East Japan Accounting Center Co.,Ltd.
 *
 * (c)2017 The University of Aizu
 * This software is released under the New BSD License.
 *
 * $Id$
 */

#ifndef ROBOTGPSDATAMANAGER_H
#define ROBOTGPSDATAMANAGER_H

#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/idl/ExtendedDataTypesSkel.h>
#include <rtm/idl/InterfaceDataTypesSkel.h>

// Service implementation headers
// <rtc-template block="service_impl_h">
#include "RobotGPSDataManagerSVC_impl.h"
#include "json.h"
#include "writer.h"
#include <chrono>


// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="consumer_stub_h">

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="port_stub_h">
// </rtc-template>

using namespace RTC;

#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

//************************************************
struct RobotGPSData_st{
	double_type dLatitude;					//緯度
	double_type dLongitude;					//経度
	double_type dAltitude;					//高さ
};

struct OutGPSDataSel_st{
	uword_type uwRobtSelect;				//GPS出力対象ロボット
	std::string strRobotName;				//ロボット名Jsonキー
	RobotGPSData_st stGPSData;				//GPSデータ
};
//************************************************

/*!
 * @class RobotGPSDataManager
 * @brief gps data management
 *
 * 別紙"機能仕様書_ロボットGPSデータ管理RTC.docx"参照
 *
 * 別紙"機能仕様書_ロボットGPSデータ管理RTC.docx"参照
 *
 * 別紙"機能仕様書_ロボットGPSデータ管理RTC.docx"参照
 *
 * 別紙"機能仕様書_ロボットGPSデータ管理RTC.docx"参照
 *
 */
class RobotGPSDataManager
  : public RTC::DataFlowComponentBase
{
 public:
  /*!
   * @brief constructor
   * @param manager Maneger Object
   */
  RobotGPSDataManager(RTC::Manager* manager);

  /*!
   * @brief destructor
   */
  ~RobotGPSDataManager();

  // <rtc-template block="public_attribute">
  
  // </rtc-template>

  // <rtc-template block="public_operation">
  
  // </rtc-template>

  /***
   * ロボットGPSデータ管理RTCの初期化処理
   *
   * The initialize action (on CREATED->ALIVE transition)
   * formaer rtc_init_entry() 
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onInitialize();

  /***
   * ロボットGPSデータ管理RTCの終了処理
   *
   * The finalize action (on ALIVE->END transition)
   * formaer rtc_exiting_entry()
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onFinalize();

  /***
   *
   * The startup action when ExecutionContext startup
   * former rtc_starting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStartup(RTC::UniqueId ec_id);

  /***
   *
   * The shutdown action when ExecutionContext stop
   * former rtc_stopping_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onShutdown(RTC::UniqueId ec_id);

  /***
   * ロボットGPSデータ管理RTCの開始処理
   *
   * The activated action (Active state entry action)
   * former rtc_active_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);

  /***
   * ロボットGPSデータ管理RTCの停止処理
   *
   * The deactivated action (Active state exit action)
   * former rtc_active_exit()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);

  /***
   * ロボットGPSデータ管理RTCの周期処理
   *
   * The execution action that is invoked periodically
   * former rtc_active_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);

  /***
   * ロボットGPSデータ管理RTCのエラー遷移処理
   *
   * The aborting action when main logic error occurred.
   * former rtc_aborting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onAborting(RTC::UniqueId ec_id);

  /***
   *
   * The error action in ERROR state
   * former rtc_error_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onError(RTC::UniqueId ec_id);

  /***
   *
   * The reset action that is invoked resetting
   * This is same but different the former rtc_init_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onReset(RTC::UniqueId ec_id);
  
  /***
   *
   * The state update action that is invoked after onExecute() action
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStateUpdate(RTC::UniqueId ec_id);

  /***
   *
   * The action that is invoked when execution context's rate is changed
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onRateChanged(RTC::UniqueId ec_id);


 protected:
  // <rtc-template block="protected_attribute">
  
  // </rtc-template>

  // <rtc-template block="protected_operation">
  
  // </rtc-template>

  // Configuration variable declaration
  // <rtc-template block="config_declare">
  /*!
   * GPSデータ出力周期を調整する
   * - Name: Output_Cycle Output_Cycle
   * - DefaultValue: 0
   * - Unit: ms
   * - Range: 0<=x<=10000
   * - Constraint: -
   */
  long int m_Output_Cycle;

  // </rtc-template>

  // DataInPort declaration
  // <rtc-template block="inport_declare">
  RTC::GPSData m_inEnryuGpsData;
  /*!
   * 援竜のGPSデータを取得するデータポート
   * - Type: RTC::GPSData
   * - Number: 1
   */
  InPort<RTC::GPSData> m_inEnryuGpsDataIn;
  RTC::GPSData m_inSpiderGpsData;
  /*!
   * スパイダーのGPSデータを取得するデータポート
   * - Type: RTC::GPSData
   * - Number: 1
   */
  InPort<RTC::GPSData> m_inSpiderGpsDataIn;
  RTC::GPSData m_inDrone1GpsData;
  /*!
   * ドローン_1のGPSデータを取得するデータポート
   * - Type: RTC::GPSData
   * - Number: 1
   */
  InPort<RTC::GPSData> m_inDrone1GpsDataIn;
  RTC::GPSData m_inDrone2GpsData;
  /*!
   * ドローン_2のGPSデータを取得するデータポート
   * - Type: RTC::GPSData
   * - Number: 1
   */
  InPort<RTC::GPSData> m_inDrone2GpsDataIn;
  RTC::GPSData m_inDrone3GpsData;
  /*!
   * ドローン_3のGPSデータを取得するデータポート
   * - Type: RTC::GPSData
   * - Number: 1
   */
  InPort<RTC::GPSData> m_inDrone3GpsDataIn;
  RTC::GPSData m_inGorone1GpsData;
  /*!
   * ゴローン1のGPSデータを取得するデータポート
   * - Type: RTC::GPSData
   * - Number: 1
   */
  InPort<RTC::GPSData> m_inGorone1GpsDataIn;
  RTC::GPSData m_inGorone2GpsData;
  /*!
   * ゴローン2のGPSデータを取得するデータポート
   * - Type: RTC::GPSData
   * - Number: 1
   */
  InPort<RTC::GPSData> m_inGorone2GpsDataIn;
  RTC::GPSData m_inGorone3GpsData;
  /*!
   * ゴローン3のGPSデータを取得するデータポート
   * - Type: RTC::GPSData
   * - Number: 1
   */
  InPort<RTC::GPSData> m_inGorone3GpsDataIn;
  RTC::GPSData m_inGorone4GpsData;
  /*!
   * ゴローン4のGPSデータを取得するデータポート
   * - Type: RTC::GPSData
   * - Number: 1
   */
  InPort<RTC::GPSData> m_inGorone4GpsDataIn;
  
  // </rtc-template>


  // DataOutPort declaration
  // <rtc-template block="outport_declare">
  RTC::TimedString m_outRobotGpsData;
  /*!
   * 纏めたロボットGPSデータをJyson形式で提供するデータポート
   * - Type: RTC::TimedString
   * - Number: 1
   * - Frequency: 500ms周期
   * - Operation Cycle: 500ms周期
   */
  OutPort<RTC::TimedString> m_outRobotGpsDataOut;
  
  // </rtc-template>

  // CORBA Port declaration
  // <rtc-template block="corbaport_declare">
  /*!
   * GPSデータ出力制御関連のインターフェース
   * Interface: GPSデータ出力制御や、対象ロボット選択などのインター
   *            フェースを提供する
   */
  RTC::CorbaPort m_GPSDataManagerReqServPort;
  
  // </rtc-template>

  // Service declaration
  // <rtc-template block="service_declare">
  /*!
   * GPSデータ出力On/Off制御要求を行う
   * - Argument:      別紙"機能仕様書_ロボットGPSデータ管理RTC.doc
   *                  x"参照
   * - Return Value:  別紙"機能仕様書_ロボットGPSデータ管理RTC.doc
   *                  x"参照
   * - Exception:     なし
   * - PreCondition:  -
   * - PostCondition: -
   */
  GPSDataOutputControlSVC_impl m_GPSDataOutputControlProv;
  /*!
   * GPSデータ出力対象ロボットの選択を行う
   * - Argument:      別紙"機能仕様書_ロボットGPSデータ管理RTC.doc
   *                  x"参照
   * - Return Value:  別紙"機能仕様書_ロボットGPSデータ管理RTC.doc
   *                  x"参照
   * - Exception:     なし
   * - PreCondition:  -
   * - PostCondition: -
   */
  GPSDataOutputSelectSVC_impl m_GPSDataOutputSelectProv;
  
  // </rtc-template>

  // Consumer declaration
  // <rtc-template block="consumer_declare">
  
  // </rtc-template>

 private:
  // <rtc-template block="private_attribute">
  
  // </rtc-template>

  // <rtc-template block="private_operation">
  
  // </rtc-template>

	 void RobotGPSDataReqSet(void);
	 uhword_type uhwRobotGPSDataOutput(void);

	 void JsonDataInit(void);
	 void GetEnryuGPSData(void);
	 void GetSpiderGPSData(void);
	 void GetDroneGPSData(void);
	 void GetGoroneGPSData(void);
	 void RobotGPSDataManagerRamInit(void);

	 uword_type m_uwCurrentOutputCnt;		//現在の出力状態
	 uword_type m_uwCurrentOutputSel;		//現在のロボット選択状態

	 Json::Value m_JsonData;				//Jsonデータ格納
	 RobotGPSData_st m_stEnryuGPSData;		//援竜GPSデータ格納
	 RobotGPSData_st m_stSpiderGPSData;		//スパイダーGPSデータ格納
	 RobotGPSData_st m_stDrone1GPSData;		//ドローン_1GPSデータ格納
	 RobotGPSData_st m_stDrone2GPSData;		//ドローン_2GPSデータ格納
	 RobotGPSData_st m_stDrone3GPSData;		//ドローン_3GPSデータ格納
	 RobotGPSData_st m_stGorone1GPSData;	//ゴローン_1GPSデータ格納
	 RobotGPSData_st m_stGorone2GPSData;	//ゴローン_2GPSデータ格納
	 RobotGPSData_st m_stGorone3GPSData;	//ゴローン_3GPSデータ格納
	 RobotGPSData_st m_stGorone4GPSData;	//ゴローン_4GPSデータ格納

	 std::chrono::system_clock::time_point  m_LastTime;	//前回のGPS出力時間を保持
};


extern "C"
{
  DLL_EXPORT void RobotGPSDataManagerInit(RTC::Manager* manager);
};

#endif // ROBOTGPSDATAMANAGER_H
