// -*- C++ -*-
/*!
 * @file  RobotGPSDataManager.cpp
 * @brief gps data management
 * @date $Date$
 *
 * @author East Japan Accounting Center Co.,Ltd.
 *
 * (c)2017 The University of Aizu
 * This software is released under the New BSD License.
 *
 * $Id$
 */

#include "RobotGPSDataManager.h"

// Module specification
// <rtc-template block="module_spec">
static const char* robotgpsdatamanager_spec[] =
  {
    "implementation_id", "RobotGPSDataManager",
    "type_name",         "RobotGPSDataManager",
    "description",       "gps data management",
    "version",           "1.0.0",
    "vendor",            "EAC",
    "category",          "GPS",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.Output_Cycle", "0",
    // Widget
    "conf.__widget__.Output_Cycle", "spin.500",
    // Constraints
    "conf.__constraints__.Output_Cycle", "0<=x<=10000",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
RobotGPSDataManager::RobotGPSDataManager(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_inEnryuGpsDataIn("inEnryuGpsData", m_inEnryuGpsData),
    m_inSpiderGpsDataIn("inSpiderGpsData", m_inSpiderGpsData),
    m_inDrone1GpsDataIn("inDrone1GpsData", m_inDrone1GpsData),
    m_inDrone2GpsDataIn("inDrone2GpsData", m_inDrone2GpsData),
    m_inDrone3GpsDataIn("inDrone3GpsData", m_inDrone3GpsData),
    m_inGorone1GpsDataIn("inGorone1GpsData", m_inGorone1GpsData),
    m_inGorone2GpsDataIn("inGorone2GpsData", m_inGorone2GpsData),
    m_inGorone3GpsDataIn("inGorone3GpsData", m_inGorone3GpsData),
    m_inGorone4GpsDataIn("inGorone4GpsData", m_inGorone4GpsData),
    m_outRobotGpsDataOut("outRobotGpsData", m_outRobotGpsData),
    m_GPSDataManagerReqServPort("GPSDataManagerReqServ")

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
RobotGPSDataManager::~RobotGPSDataManager()
{
}


/*!
 * ロボットGPSデータ管理RTCの初期化処理
 */
RTC::ReturnCode_t RobotGPSDataManager::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("inEnryuGpsData", m_inEnryuGpsDataIn);
  addInPort("inSpiderGpsData", m_inSpiderGpsDataIn);
  addInPort("inDrone1GpsData", m_inDrone1GpsDataIn);
  addInPort("inDrone2GpsData", m_inDrone2GpsDataIn);
  addInPort("inDrone3GpsData", m_inDrone3GpsDataIn);
  addInPort("inGorone1GpsData", m_inGorone1GpsDataIn);
  addInPort("inGorone2GpsData", m_inGorone2GpsDataIn);
  addInPort("inGorone3GpsData", m_inGorone3GpsDataIn);
  addInPort("inGorone4GpsData", m_inGorone4GpsDataIn);
  
  // Set OutPort buffer
  addOutPort("outRobotGpsData", m_outRobotGpsDataOut);
  
  // Set service provider to Ports
  m_GPSDataManagerReqServPort.registerProvider("GPSDataOutputControl", "GPSDataManagerReqServ::GPSDataOutputControl", m_GPSDataOutputControlProv);
  m_GPSDataManagerReqServPort.registerProvider("GPSDataOutputSelect", "GPSDataManagerReqServ::GPSDataOutputSelect", m_GPSDataOutputSelectProv);
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  addPort(m_GPSDataManagerReqServPort);
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("Output_Cycle", m_Output_Cycle, "0");
  // </rtc-template>

#ifdef _DEBUG
  std::cout << "[info]RobotGPSDataManager::onInitialize()" << std::endl;
#endif
  RobotGPSDataManagerRamInit();

  return RTC::RTC_OK;
}

/*!
 * ロボットGPSデータ管理RTCの終了処理
 */

RTC::ReturnCode_t RobotGPSDataManager::onFinalize()
{
#ifdef _DEBUG
	std::cout << "[info]RobotGPSDataManager::onFinalize()" << std::endl;
#endif

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RobotGPSDataManager::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RobotGPSDataManager::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * ロボットGPSデータ管理RTCの開始処理
 */

RTC::ReturnCode_t RobotGPSDataManager::onActivated(RTC::UniqueId ec_id)
{
#ifdef _DEBUG
	std::cout << "[info]RobotGPSDataManager::onActivated()" << std::endl;
#endif

	//入力データポートの初期化
	while(m_inEnryuGpsDataIn.isEmpty() == FALSE){
		m_inEnryuGpsDataIn.read();
	}
	while(m_inSpiderGpsDataIn.isEmpty() == FALSE){
		m_inSpiderGpsDataIn.read();
	}
	while(m_inDrone1GpsDataIn.isEmpty() == FALSE){
		m_inDrone1GpsDataIn.read();
	}
	while(m_inDrone2GpsDataIn.isEmpty() == FALSE){
		m_inDrone2GpsDataIn.read();
	}
	while(m_inDrone3GpsDataIn.isEmpty() == FALSE){
		m_inDrone3GpsDataIn.read();
	}
	while(m_inGorone1GpsDataIn.isEmpty() == FALSE){
		m_inGorone1GpsDataIn.read();
	}
	while(m_inGorone2GpsDataIn.isEmpty() == FALSE){
		m_inGorone2GpsDataIn.read();
	}
	while(m_inGorone3GpsDataIn.isEmpty() == FALSE){
		m_inGorone3GpsDataIn.read();
	}
	while(m_inGorone4GpsDataIn.isEmpty() == FALSE){
		m_inGorone4GpsDataIn.read();
	}

	//経過時間保持変数を初期化
	m_LastTime = std::chrono::system_clock::now();

	return RTC::RTC_OK;
}

/*!
 * ロボットGPSデータ管理RTCの停止処理
 */

RTC::ReturnCode_t RobotGPSDataManager::onDeactivated(RTC::UniqueId ec_id)
{
#ifdef _DEBUG
	std::cout << "[info]RobotGPSDataManager::onDeactivated()" << std::endl;
#endif
	RobotGPSDataManagerRamInit();

	return RTC::RTC_OK;
}

/*!
 * ロボットGPSデータ管理RTCの周期処理
 */

RTC::ReturnCode_t RobotGPSDataManager::onExecute(RTC::UniqueId ec_id)
{
	std::chrono::system_clock::time_point  CurrentTime; 
	double dElapsedTime = 0;

	CurrentTime = std::chrono::system_clock::now();	//現在の時間を取得
	dElapsedTime = std::chrono::duration_cast<std::chrono::milliseconds>(CurrentTime - m_LastTime).count();	//経過時間をミリ秒に変換

	//GPSデータ取得
	GetEnryuGPSData();
	GetSpiderGPSData();
	GetDroneGPSData();
	GetGoroneGPSData();

	//コンフィギュレーション出力周期経過後、GPSデータ出力
	if(dElapsedTime >= m_Output_Cycle){
		uhwRobotGPSDataOutput();
		m_LastTime = CurrentTime;
	}

	return RTC::RTC_OK;
}

/*!
 * ロボットGPSデータ管理RTCのエラー遷移処理
 */

RTC::ReturnCode_t RobotGPSDataManager::onAborting(RTC::UniqueId ec_id)
{
#ifdef _DEBUG
	std::cout << "[info]RobotGPSDataManager::onAborting()" << std::endl;
#endif

	return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RobotGPSDataManager::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RobotGPSDataManager::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RobotGPSDataManager::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RobotGPSDataManager::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*******************************************************************************************
	サブルーチン
*******************************************************************************************/

/******************************************************************************************
*	Function Name		：RobotGPSDataReqSet
*	Description			：ロボットGPS出力要求設定
*	Argument			：なし
*	Return				：なし
******************************************************************************************/
void RobotGPSDataManager::RobotGPSDataReqSet(void)
{
	uword_type uwOutputCnt = GPS_DISP_OFF;
	uword_type uwOutputSel = ROBOT_SEL_BIT_OFF;

	m_GPSDataOutputControlProv.GetGPSDataOutputControl(uwOutputCnt);
	if(uwOutputCnt >= GPS_DISP_MAX){
		std::cout << "[warrning]GetGPSDataOutputControl is range out!!" << std::endl;
		return;
	}
	//変化チェック
	if(uwOutputCnt != m_uwCurrentOutputCnt){
		std::cout << "[debug]GetGPSDataOutputControl = " << uwOutputCnt <<std::endl;
		m_uwCurrentOutputCnt = uwOutputCnt;
	}

	m_GPSDataOutputSelectProv.GetGPSDataOutputSelect(uwOutputSel);
	if(uwOutputSel >= ROBOT_SEL_BIT_MAX){
		std::cout << "[warrning]GetGPSDataOutputSelect is range out!!" << std::endl;
		return;
	}
	//変化チェック
	if(uwOutputSel != m_uwCurrentOutputSel){
		std::cout << "[debug]GetGPSDataOutputSelect = " << uwOutputSel <<std::endl;
		m_uwCurrentOutputSel = uwOutputSel;
	}
}

/******************************************************************************************
*	Function Name		：GetEnryuGPSData
*	Description			：データポートから援竜GPSデータを取得する
*	Argument			：なし
*	Return				：なし
******************************************************************************************/
void RobotGPSDataManager::GetEnryuGPSData(void)
{
	if(m_inEnryuGpsDataIn.isNew()){
		m_inEnryuGpsDataIn.read();
		m_stEnryuGPSData.dLatitude = m_inEnryuGpsData.latitude;
		m_stEnryuGPSData.dLongitude = m_inEnryuGpsData.longitude;
		m_stEnryuGPSData.dAltitude = m_inEnryuGpsData.altitude;
	}
}

/******************************************************************************************
*	Function Name		：GetSpiderGPSData
*	Description			：データポートからスパイダーGPSデータを取得する
*	Argument			：なし
*	Return				：なし
******************************************************************************************/
void RobotGPSDataManager::GetSpiderGPSData(void)
{
	if(m_inSpiderGpsDataIn.isNew()){
		m_inSpiderGpsDataIn.read();
		m_stSpiderGPSData.dLatitude = m_inSpiderGpsData.latitude;
		m_stSpiderGPSData.dLongitude = m_inSpiderGpsData.longitude;
		m_stSpiderGPSData.dAltitude = m_inSpiderGpsData.altitude;
	}
}

/******************************************************************************************
*	Function Name		：GetDroneGPSData
*	Description			：データポートからドローンGPSデータを取得する
*	Argument			：なし
*	Return				：なし
******************************************************************************************/
void RobotGPSDataManager::GetDroneGPSData(void)
{
	if(m_inDrone1GpsDataIn.isNew()){
		m_inDrone1GpsDataIn.read();
		m_stDrone1GPSData.dLatitude = m_inDrone1GpsData.latitude;
		m_stDrone1GPSData.dLongitude = m_inDrone1GpsData.longitude;
		m_stDrone1GPSData.dAltitude = m_inDrone1GpsData.altitude;
	}

	if(m_inDrone2GpsDataIn.isNew()){
		m_inDrone2GpsDataIn.read();
		m_stDrone2GPSData.dLatitude = m_inDrone2GpsData.latitude;
		m_stDrone2GPSData.dLongitude = m_inDrone2GpsData.longitude;
		m_stDrone2GPSData.dAltitude = m_inDrone2GpsData.altitude;
	}

	if(m_inDrone3GpsDataIn.isNew()){
		m_inDrone3GpsDataIn.read();
		m_stDrone3GPSData.dLatitude = m_inDrone3GpsData.latitude;
		m_stDrone3GPSData.dLongitude = m_inDrone3GpsData.longitude;
		m_stDrone3GPSData.dAltitude = m_inDrone3GpsData.altitude;
	}
}

/******************************************************************************************
*	Function Name		：GetGoroneGPSData
*	Description			：データポートからゴローンGPSデータを取得する
*	Argument			：なし
*	Return				：なし
******************************************************************************************/
void RobotGPSDataManager::GetGoroneGPSData(void)
{
	if(m_inGorone1GpsDataIn.isNew()){
		m_inGorone1GpsDataIn.read();
		m_stGorone1GPSData.dLatitude = m_inGorone1GpsData.latitude;
		m_stGorone1GPSData.dLongitude = m_inGorone1GpsData.longitude;
		m_stGorone1GPSData.dAltitude = m_inGorone1GpsData.altitude;
	}

	if(m_inGorone2GpsDataIn.isNew()){
		m_inGorone2GpsDataIn.read();
		m_stGorone2GPSData.dLatitude = m_inGorone2GpsData.latitude;
		m_stGorone2GPSData.dLongitude = m_inGorone2GpsData.longitude;
		m_stGorone2GPSData.dAltitude = m_inGorone2GpsData.altitude;
	}

	if(m_inGorone3GpsDataIn.isNew()){
		m_inGorone3GpsDataIn.read();
		m_stGorone3GPSData.dLatitude = m_inGorone3GpsData.latitude;
		m_stGorone3GPSData.dLongitude = m_inGorone3GpsData.longitude;
		m_stGorone3GPSData.dAltitude = m_inGorone3GpsData.altitude;
	}

	if(m_inGorone4GpsDataIn.isNew()){
		m_inGorone4GpsDataIn.read();
		m_stGorone4GPSData.dLatitude = m_inGorone4GpsData.latitude;
		m_stGorone4GPSData.dLongitude = m_inGorone4GpsData.longitude;
		m_stGorone4GPSData.dAltitude = m_inGorone4GpsData.altitude;
	}
}

/******************************************************************************************
*	Function Name		：JsonDataInit
*	Description			：Jsonデータ格納変数の初期化
*	Argument			：なし
*	Return				：なし
******************************************************************************************/
void RobotGPSDataManager::JsonDataInit(void)
{
	//Jsonデータクリア
	m_JsonData.clear();

	m_JsonData[JSONKEY_ROBOT_ENRYU][JSONKEY_LAT] = 0;
	m_JsonData[JSONKEY_ROBOT_ENRYU][JSONKEY_LON] = 0;
	m_JsonData[JSONKEY_ROBOT_ENRYU][JSONKEY_ALT] = 0;

	m_JsonData[JSONKEY_ROBOT_SPIDER][JSONKEY_LAT] = 0;
	m_JsonData[JSONKEY_ROBOT_SPIDER][JSONKEY_LON] = 0;
	m_JsonData[JSONKEY_ROBOT_SPIDER][JSONKEY_ALT] = 0;

	m_JsonData[JSONKEY_ROBOT_DRONE_1][JSONKEY_LAT] = 0;
	m_JsonData[JSONKEY_ROBOT_DRONE_1][JSONKEY_LON] = 0;
	m_JsonData[JSONKEY_ROBOT_DRONE_1][JSONKEY_ALT] = 0;
	m_JsonData[JSONKEY_ROBOT_DRONE_2][JSONKEY_LAT] = 0;
	m_JsonData[JSONKEY_ROBOT_DRONE_2][JSONKEY_LON] = 0;
	m_JsonData[JSONKEY_ROBOT_DRONE_2][JSONKEY_ALT] = 0;
	m_JsonData[JSONKEY_ROBOT_DRONE_3][JSONKEY_LAT] = 0;
	m_JsonData[JSONKEY_ROBOT_DRONE_3][JSONKEY_LON] = 0;
	m_JsonData[JSONKEY_ROBOT_DRONE_3][JSONKEY_ALT] = 0;

	m_JsonData[JSONKEY_ROBOT_GORONE_1][JSONKEY_LAT] = 0;
	m_JsonData[JSONKEY_ROBOT_GORONE_1][JSONKEY_LON] = 0;
	m_JsonData[JSONKEY_ROBOT_GORONE_1][JSONKEY_ALT] = 0;
	m_JsonData[JSONKEY_ROBOT_GORONE_2][JSONKEY_LAT] = 0;
	m_JsonData[JSONKEY_ROBOT_GORONE_2][JSONKEY_LON] = 0;
	m_JsonData[JSONKEY_ROBOT_GORONE_2][JSONKEY_ALT] = 0;
	m_JsonData[JSONKEY_ROBOT_GORONE_3][JSONKEY_LAT] = 0;
	m_JsonData[JSONKEY_ROBOT_GORONE_3][JSONKEY_LON] = 0;
	m_JsonData[JSONKEY_ROBOT_GORONE_3][JSONKEY_ALT] = 0;
	m_JsonData[JSONKEY_ROBOT_GORONE_4][JSONKEY_LAT] = 0;
	m_JsonData[JSONKEY_ROBOT_GORONE_4][JSONKEY_LON] = 0;
	m_JsonData[JSONKEY_ROBOT_GORONE_4][JSONKEY_ALT] = 0;
}

/******************************************************************************************
*	Function Name		：uhwRobotGPSDataOutput
*	Description			：GPSデータ出力
*	Argument			：なし
*	Return				：なし
******************************************************************************************/
uhword_type RobotGPSDataManager::uhwRobotGPSDataOutput(void)
{
	Json::StyledWriter	writer;
	std::string strJson = "";

	struct OutGPSDataSel_st OutGPSDataSelectTbl[9] ={
		{ROBOT_SEL_BIT_ENRYU,	JSONKEY_ROBOT_ENRYU,	m_stEnryuGPSData },
		{ROBOT_SEL_BIT_SPIDER,	JSONKEY_ROBOT_SPIDER,	m_stSpiderGPSData},
		{ROBOT_SEL_BIT_DRONE_1,	JSONKEY_ROBOT_DRONE_1,	m_stDrone1GPSData},
		{ROBOT_SEL_BIT_DRONE_2,	JSONKEY_ROBOT_DRONE_2,	m_stDrone2GPSData},
		{ROBOT_SEL_BIT_DRONE_3,	JSONKEY_ROBOT_DRONE_3,	m_stDrone3GPSData},
		{ROBOT_SEL_BIT_GORONE_1,JSONKEY_ROBOT_GORONE_1,	m_stGorone1GPSData},
		{ROBOT_SEL_BIT_GORONE_2,JSONKEY_ROBOT_GORONE_2,	m_stGorone2GPSData},
		{ROBOT_SEL_BIT_GORONE_3,JSONKEY_ROBOT_GORONE_3,	m_stGorone3GPSData},
		{ROBOT_SEL_BIT_GORONE_4,JSONKEY_ROBOT_GORONE_4,	m_stGorone4GPSData}
	};

	RobotGPSDataReqSet();

	if(m_uwCurrentOutputCnt == GPS_DISP_ON){
		if(m_uwCurrentOutputSel == ROBOT_SEL_BIT_OFF){
			std::cout << "[warring]GPS output request parameter is undefined!! = " << 
				m_uwCurrentOutputCnt << ", " << m_uwCurrentOutputSel << std::endl;
			return RESULT_WARNING;
		}else{
			//JSON格納変数の初期化
			JsonDataInit();

			//データ作成
			for(uhword_type uhwLoop = 0; uhwLoop < 9; uhwLoop ++){
				if(m_uwCurrentOutputSel & OutGPSDataSelectTbl[uhwLoop].uwRobtSelect){
					std::string strRobotNameKey = OutGPSDataSelectTbl[uhwLoop].strRobotName;
					m_JsonData[strRobotNameKey][JSONKEY_LAT] = OutGPSDataSelectTbl[uhwLoop].stGPSData.dLatitude;
					m_JsonData[strRobotNameKey][JSONKEY_LON] = OutGPSDataSelectTbl[uhwLoop].stGPSData.dLongitude;
					m_JsonData[strRobotNameKey][JSONKEY_ALT] = OutGPSDataSelectTbl[uhwLoop].stGPSData.dAltitude;
				}
			}
			//jsonフォーマットに変換し出力
			strJson = writer.write(m_JsonData);

			std::cout << "[debug]output = " << strJson << std::endl;

			m_outRobotGpsData.data = strJson.c_str();
			m_outRobotGpsDataOut.write();
		}
	}
	
	return RESULT_SUCCESS;
}

/******************************************************************************************
*	Function Name		：RobotGPSDataManagerRamInit
*	Description			：使用変数初期化処理
*	Argument			：なし
*	Return				：なし
******************************************************************************************/
void RobotGPSDataManager::RobotGPSDataManagerRamInit(void)
{
	m_uwCurrentOutputCnt = GPS_DISP_OFF;
	m_uwCurrentOutputSel = ROBOT_SEL_BIT_OFF;
	memset(&m_stEnryuGPSData,0,sizeof(RobotGPSData_st));
	memset(&m_stSpiderGPSData,0,sizeof(RobotGPSData_st));
	memset(&m_stDrone1GPSData,0,sizeof(RobotGPSData_st));
	memset(&m_stDrone2GPSData,0,sizeof(RobotGPSData_st));
	memset(&m_stDrone3GPSData,0,sizeof(RobotGPSData_st));
	memset(&m_stGorone1GPSData,0,sizeof(RobotGPSData_st));
	memset(&m_stGorone2GPSData,0,sizeof(RobotGPSData_st));
	memset(&m_stGorone3GPSData,0,sizeof(RobotGPSData_st));
	memset(&m_stGorone4GPSData,0,sizeof(RobotGPSData_st));

	m_GPSDataOutputControlProv.initGPSDataOutputControl();
	m_GPSDataOutputSelectProv.initGPSDataOutputSelect();
}

extern "C"
{
 
  void RobotGPSDataManagerInit(RTC::Manager* manager)
  {
    coil::Properties profile(robotgpsdatamanager_spec);
    manager->registerFactory(profile,
                             RTC::Create<RobotGPSDataManager>,
                             RTC::Delete<RobotGPSDataManager>);
  }
  
};


